import io
import os
import re
import subprocess
import tempfile
from contextlib import contextmanager
from os.path import abspath, dirname
from typing import Tuple, Union

import click
import requests
import yaml
import yamlinclude
from mach import exceptions, git
from yaml.error import YAMLError  # noqa

EXTERNAL_RE = re.compile(r"^(git::)?(http|https)://")
INCLUDE_RE = re.compile(r"^(.*)\${include\((.*)\)}\s*$")

Encrypted = bool

__all__ = ["YAMLError", "load"]


class YamlFileIO(io.TextIOWrapper):
    """YAML IO wrapper to modify the yaml input stream before parsing it.

    We support the !include tag using the yamlinclude package.
    However, MACH will provide this option in the same fashion as variables can be defined
    using the `${...}` syntax: `${include(file.yml)}`.
    An added bonus here is that SOPS will not strip this syntax as it does with !include
    """

    def read(self, size=-1):
        lines = super().read(size)
        return "\n".join([self.parse_line(line) for line in lines.split("\n")])

    def parse_line(self, line: str) -> str:
        match = INCLUDE_RE.match(line)
        if match:
            prefix, include_part = match.groups()
            line = f"{prefix}!include {include_part}"
        return line


def load(file: str) -> Tuple[dict, Encrypted]:
    YamlIncludeConstructor.add_to_loader_class(
        loader_class=yaml.FullLoader,
        base_dir=abspath(dirname(file)),
    )
    encrypted = False

    with open(file, "r+b") as fh:
        data = _yaml_load(fh)

    if "sops" in data:
        click.echo("Detected SOPS encryption; decrypting...")
        data = _yaml_load(_sops_stream(file))
        encrypted = True

    return data, encrypted


def _yaml_load(iostream: io.IOBase):
    yaml_io = YamlFileIO(iostream)
    return yaml.full_load(yaml_io)


def _sops_stream(file: str, *args, **kwargs) -> bytes:
    kwargs["stderr"] = subprocess.STDOUT
    cmd = ["sops", "-d", file, "--output-type=yaml"]
    try:
        return io.BytesIO(subprocess.check_output(cmd, *args, **kwargs))
    except subprocess.CalledProcessError as e:
        raise exceptions.MachError(e.output.decode() if e.output else str(e)) from e


class YamlIncludeConstructor(yamlinclude.YamlIncludeConstructor):
    def load(
        self,
        loader: Union[yaml.Loader, yaml.FullLoader],
        pathname: str,
        recursive: bool = False,
        encoding: str = "",
        reader: str = "",
    ):
        with resolve_file(pathname) as file:
            return super().load(loader, file, recursive, encoding, reader)


@contextmanager
def resolve_file(path):
    if not EXTERNAL_RE.match(path):
        yield path
        return

    git_match = re.match(r"^git::(.*)", path)
    if git_match:
        git_path = git_match.group(1)
        with tempfile.TemporaryDirectory() as tmpdir:
            yield resolve_git(git_path, tmpdir)
    elif path.startswith("http"):
        suffix = "." + path.rsplit(".", 1)[1]
        with tempfile.NamedTemporaryFile(suffix=suffix) as tmpfile:
            yield resolve_http(path, tmpfile)
    else:
        raise Exception(f"External path {path} not supported.")


def resolve_http(path: str, tmpfile: io.FileIO):
    resp = requests.get(path)
    resp.raise_for_status()
    tmpfile.write(resp.content)
    tmpfile.seek(0)
    return tmpfile.name


def resolve_git(path: str, tmpdir: str):
    """Resolve file served on Git.

    Return path of local file
    """
    match = re.match(r"(.*\/.*)\/\/(.*)$", path)
    if match:
        repo, file = match.groups()
    else:
        raise Exception(
            f"Missing file component in include {path}: can't be just the Git repository"
        )

    click.echo(f"Resolving {file} from {repo}")
    dest = os.path.join(tmpdir, "repo")

    try:
        git.ensure_local(repo, dest)
    except git.GitError as e:
        raise exceptions.MachError(f"Could not fetch {repo}: {e}") from e

    return os.path.join(dest, file)



from dataclasses import dataclass
from typing import Optional

from dataclasses_json import dataclass_json
from dataclasses_jsonschema import JsonSchemaMixin

from . import fields

__all__ = ["ComponentAzureConfig", "ServicePlan"]


@dataclass_json
@dataclass
class ComponentAzureConfig(JsonSchemaMixin):
    service_plan: Optional[str] = fields.none()
    short_name: Optional[str] = fields.none()

    def merge(self, config: "ComponentAzureConfig"):
        self.service_plan = self.service_plan or config.service_plan
        self.short_name = self.short_name or config.short_name


@dataclass_json
@dataclass
class ServicePlan(JsonSchemaMixin):
    kind: str
    tier: str
    size: str
    capacity: Optional[int] = fields.none()
    dedicated_resource_group: bool = fields.default(False)
    per_site_scaling: bool = fields.default(False)



import re
import textwrap
import warnings
from collections import defaultdict
from dataclasses import dataclass
from os.path import abspath, basename, splitext
from pathlib import Path
from typing import Any, Dict, List, Tuple

import click
from mach import exceptions, yaml
from mach.types import (
    CloudOption,
    ComponentAzureConfig,
    ComponentConfig,
    Endpoint,
    MachConfig,
    SentryDsn,
    ServicePlan,
    Site,
    SiteAzureSettings,
    TerraformReference,
)
from mach.validate import validate_config
from mach.variables import resolve_env_variable, resolve_variable
from marshmallow.exceptions import ValidationError

VARIABLE_RE = re.compile(r"^\${(var|env)\.(.*)}$")


def parse_components(file: str):
    yaml_data, _ = yaml.load(file)

    try:
        with warnings.catch_warnings():
            # Suppress a 'Unknown type ForwardRef('Component')' warning from dataclasses_json
            warnings.simplefilter("ignore")
            components = ComponentConfig.schema(infer_missing=True, many=True).load(
                yaml_data
            )
    except KeyError as e:
        # Most probably a missing value in the configuration.
        # dataclasses_json doesn't really give a proper Exception for this.
        # TODO: See if we can improve this / make it more robust. Either by improving
        # dataclassess_json (with a PR) or by extending it (if possible)
        raise exceptions.ParseError(f"Required attribute {e} missing") from e
    except ValidationError as e:
        # TODO: We don't have any path here, so not the best of error messages
        raise exceptions.ParseError(
            "Configuration file could not be validated", details=e.normalized_messages()
        ) from e

    return components


@dataclass
class VariablesFile:
    path: str
    vars: dict
    encrypted: bool

    @classmethod
    def from_file(cls, path: str):
        try:
            vars, vars_encrypted = yaml.load(path)
        except yaml.YAMLError as e:
            raise exceptions.ParseError(f"Could not parse variables file:\n{e}")

        return cls(path=path, vars=vars, encrypted=vars_encrypted)


def parse_configs(
    files: List[str],
    output_path: str = None,
    *,
    ignore_version=True,
    var_file: str = "",
) -> List[MachConfig]:
    """Parse and validate configurations."""
    vars = VariablesFile.from_file(var_file) if var_file else None

    configs = []
    for file in files:
        try:
            config = parse_and_validate(
                file,
                output_path,
                ignore_version=ignore_version,
                vars=vars,
            )
        except exceptions.ParseError as e:
            click.echo(textwrap.indent(str(e), "  "))
            continue

        configs.append(config)

    return configs


def parse_and_validate(
    file: str,
    output_path: str = None,
    *,
    ignore_version=True,
    vars: VariablesFile = None,
    vars_encrypted=False,
) -> MachConfig:
    """Parse and validate configuration."""
    yaml_content, encrypted = yaml.load(file)
    if not vars:
        vars = get_config_vars(yaml_content)

    config = parse_config_from_yaml(yaml_content, vars=vars)
    config.file_encrypted = encrypted
    config.file = file

    validate_config(config, ignore_version=ignore_version)

    if output_path:
        full_output_path = Path(f"{output_path}/{splitext(basename(file))[0]}")
        full_output_path.mkdir(exist_ok=True, parents=True)
        config.output_path = str(full_output_path)

    return config


def get_config_vars(yaml_content: dict) -> VariablesFile:
    var_file = yaml_content.get("mach_composer", {}).get("variables_file", None)
    if not var_file:
        return None

    return VariablesFile.from_file(var_file)


def parse_config_from_file(file: str, *, vars: VariablesFile = None) -> MachConfig:
    click.echo(f"Parsing {file}...")
    yaml_content, encrypted = yaml.load(file)
    config = parse_config_from_yaml(yaml_content, vars=vars)
    config.file_encrypted = encrypted
    config.file = file
    return config


def parse_config_from_yaml(
    yaml_content: dict, *, vars: VariablesFile = None
) -> MachConfig:
    """Parse file into MachConfig object."""
    try:
        with warnings.catch_warnings():
            # Suppress a 'Unknown type ForwardRef('Component')' warning from dataclasses_json
            warnings.simplefilter("ignore")
            config = MachConfig.schema(infer_missing=True).load(yaml_content)  # type: ignore

        if vars:
            config.variables = vars.vars
            config.variables_encrypted = vars.encrypted
            config.variables_path = vars.path

    except KeyError as e:
        # Most probably a missing value in the configuration.
        # dataclasses_json doesn't really give a proper Exception for this.
        # TODO: See if we can improve this / make it more robust. Either by improving
        # dataclassess_json (with a PR) or by extending it (if possible)
        raise exceptions.ParseError(f"Required attribute {e} missing") from e
    except ValidationError as e:
        # TODO: We don't have any path here, so not the best of error messages
        raise exceptions.ParseError(
            "Configuration file could not be validated", details=e.normalized_messages()
        ) from e

    return parse_config(config)


def parse_config(config: MachConfig) -> MachConfig:
    resolve_variables(config, config.variables, config.variables_encrypted)
    parse_global_config(config)
    resolve_component_definitions(config)
    resolve_site_configs(config)
    return config


def resolve_variables(obj: Any, vars: dict, vars_encrypted: bool = False):
    """Resolve variables in the configuration.

    Only look for ${var.} variables since these can and must be rendered
    during parsing phase.

    This method will loop over objects that are either;
    - a dataclass (all configuration types are dataclasses)
    - a list of values
    - a dictionary

    ${component.} variables for example must be rendered in the Jinja template.
    """
    if isinstance(obj, str):
        var_m = VARIABLE_RE.match(obj.strip())
        if not var_m:
            # Return as is.
            return obj

        type_, var_name = var_m.groups()

        if type_ == "var":
            # We'll resolve the variable.
            # In case of encrypted vars we won't return the value as is,
            # but still the function will raise an error in case the variable cannot be found.
            # Better to catch it here then during terraform apply.
            var_value = resolve_variable(var_name, vars)

            if vars_encrypted:
                return TerraformReference(
                    f'data.sops_external.variables.data["{var_name}"]'
                )
            return var_value
        elif type_ == "env":
            return resolve_env_variable(var_name)
        else:
            raise exceptions.MachError(f"Unsupported variables type '{type_}': {obj}")

    annotations = getattr(obj, "__annotations__", {})
    if annotations:
        for field in annotations.keys():
            value = getattr(obj, field, None)
            if not value:
                continue
            value = resolve_variables(value, vars, vars_encrypted)
            setattr(obj, field, value)
    elif isinstance(obj, list):
        return [resolve_variables(v, vars, vars_encrypted) for v in obj]
    elif isinstance(obj, dict):
        return {k: resolve_variables(v, vars, vars_encrypted) for k, v in obj.items()}

    return obj


def parse_global_config(config: MachConfig):
    if config.general_config.cloud == CloudOption.AZURE:
        assert config.general_config.azure
        if "default" not in config.general_config.azure.service_plans:
            config.general_config.azure.service_plans["default"] = ServicePlan(
                kind="FunctionApp", tier="Dynamic", size="Y1"
            )


# flake8: noqa: C901
def resolve_site_configs(config: MachConfig):
    """Use and merge site-specific configurations with general config."""
    for site in config.sites:
        if config.general_config.cloud == CloudOption.AZURE:
            assert config.general_config.azure

            if site.azure:
                site.azure.merge(config.general_config.azure)
            else:
                site.azure = SiteAzureSettings.from_config(config.general_config.azure)

            assert site.azure

            if site.azure.resource_group:
                click.echo(
                    click.style(
                        (
                            f"WARNING: resource_group on {site.identifier} "
                            f"is used ({site.azure.resource_group}). "
                        ),
                        fg="red",
                        bold=True,
                    )
                )
                click.echo(
                    click.style(
                        (
                            "   Make sure it wasn't managed by MACH before otherwise "
                            "the resource group will get deleted."
                        ),
                        fg="red",
                    )
                )

    # Merge Contentful settings
    if config.general_config.contentful:
        for site in config.sites:
            if site.contentful:
                site.contentful.merge(config.general_config.contentful)

    # Merge Amplience settings
    if config.general_config.amplience:
        for site in config.sites:
            if site.amplience:
                site.amplience.merge(config.general_config.amplience)

    if config.general_config.sentry:
        for site in config.sites:
            if not site.sentry:
                site.sentry = SentryDsn.from_config(config.general_config.sentry)
            else:
                site.sentry.merge(config.general_config.sentry)

    resolve_site_components(config)

    for site in config.sites:
        resolve_endpoint_components(site)
        if site.azure:
            resolve_used_service_plans(site)


def resolve_used_service_plans(site: Site):
    "Azure-specific method to find out which service plans are actually being used by components"
    assert site.azure

    used = [
        c.azure.service_plan
        for c in site.components
        if c.azure and c.azure.service_plan
    ]
    site.azure.service_plans = {
        key: sp for key, sp in site.azure.service_plans.items() if key in used
    }


def resolve_endpoint_components(site: Site):
    endpoint_components = defaultdict(list)
    for c in site.components:
        for endpoint_key in c.endpoints.values():
            endpoint_components[endpoint_key].append(c)

    site_endpoint_keys = {e.key for e in site.endpoints}
    # If one of the components has a 'default' endpoint defined,
    # we'll include it to our site endpoints.
    # A 'default' endpoint is one without a custom domain, so no further
    # Route53 or DNS zone settings required.
    if "default" in endpoint_components and not "default" in site_endpoint_keys:
        click.echo(
            click.style(
                (
                    "WARNING: 'default' endpoint used but not defined in the site endpoints.\n"
                    "MACH will create a default endpoint without any custom domain attached to it.\n"
                    "More info: https://docs.machcomposer.io/reference/syntax/sites.html#endpoints"
                ),
                fg="yellow",
            )
        )
        site.endpoints.append(
            Endpoint(
                url="",
                key="default",
            )
        )

    for endpoint in site.endpoints:
        endpoint.components = endpoint_components.get(endpoint.key, [])


def resolve_site_components(config: MachConfig) -> MachConfig:
    """If no component info is specified, use global component settings."""
    component_info: Dict[str, ComponentConfig] = {
        component.name: component for component in config.components
    }
    for site in config.sites:
        if not site.components:
            continue

        for component in site.components:
            try:
                info = component_info[component.name]
            except KeyError:
                raise exceptions.ParseError(
                    f"Component {component.name} does not exist in global components."
                )
            component.definition = info

            if site.sentry:
                if not component.sentry:
                    component.sentry = SentryDsn.from_config(site.sentry)
                else:
                    component.sentry.merge(site.sentry)

            if site.azure:
                if not component.azure:
                    component.azure = info.azure
                elif info.azure:
                    component.azure.merge(info.azure)

    return config


def resolve_component_definitions(config: MachConfig):
    for comp in config.components:
        # Terraform needs absolute paths to modules
        if comp.source.startswith("."):
            comp.source = abspath(comp.source)

        if not comp.integrations:
            # If no integrations are given, set the Cloud integrations as default
            if config.general_config.cloud == CloudOption.AWS:
                comp.integrations = ["aws"]
            elif config.general_config.cloud == CloudOption.AZURE:
                comp.integrations = ["azure"]

        if config.general_config.cloud == CloudOption.AZURE:
            if not comp.azure:
                comp.azure = ComponentAzureConfig()
            if not comp.azure.short_name:
                comp.azure.short_name = comp.name



import os
import re
from typing import List

from jinja2 import Environment, FileSystemLoader
from jinja2.filters import do_mark_safe
from mach import utils
from mach.exceptions import MachError
from mach.types.sites import Component, Endpoint
from mach.types.values import TerraformReference

VARIABLE_RE = re.compile(r"^\${(component|var)\.(.*)}$")


def setup_jinja() -> Environment:
    templates_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)))
    env = Environment(
        loader=FileSystemLoader(templates_dir), trim_blocks=True, lstrip_blocks=True
    )
    load_filters(env)
    return env


def load_filters(env: Environment):
    env.filters.update(
        {
            "tfvalue": render_tfvalue,
            # Alias:
            "tf": render_tfvalue,
            "azure_region_long": azure_region_long,
            "azure_region_short": azure_region_short,
            "azure_frontend_endpoint_name": azure_frontend_endpoint_name,
            "zone_name": zone_name,
            "slugify": utils.slugify,
            "service_plan_resource_name": service_plan_resource_name,
            "render_commercetools_scopes": render_commercetools_scopes,
            "component_endpoint_name": component_endpoint_name,
        }
    )


def render_tfvalue(value):
    parsed_variable = parse_config_variable(value)
    if parsed_variable:
        return do_mark_safe(parsed_variable)

    if isinstance(value, TerraformReference):
        return value
    if isinstance(value, bool):
        return "true" if value else "false"
    if isinstance(value, (int, float)):
        return value
    if isinstance(value, list):
        values = ",".join([render_tfvalue(val) for val in value])
        return f"[{values}]"
    if isinstance(value, dict):
        values = "\n".join(
            [f"{key} = {render_tfvalue(val)}" for key, val in value.items()]
        )
        return f"{{\n{values}\n}}"
    return do_mark_safe(f'"{value}"')


def parse_config_variable(value: any):
    var_m = VARIABLE_RE.match(str(value).strip())
    if not var_m:
        return None

    type_, var = var_m.groups()
    if type_ == "component":
        parts = var.split(".", maxsplit=1)
        if len(parts) < 2:
            raise MachError(
                f"Invalid variable '{value}'; "
                "When using a ${component...} variable it has to consist of 2 parts; "
                "component-name.output-name"
            )
        return f"module.{parts[0]}.{parts[1]}"
    if type_ == "var":
        # This should not happen
        raise MachError(
            "${var...} variables should have been parsed already by MACH composer. "
            "Something went wrong."
        )

    return None


AZURE_REGION_DISPLAY_MAP_LONG = {
    "eastasia": "East Asia",
    "southeastasia": "Southeast Asia",
    "centralus": "Central US",
    "eastus": "East US",
    "eastus2": "East US 2",
    "westus": "West US",
    "northcentralus": "North Central US",
    "southcentralus": "South Central US",
    "northeurope": "North Europe",
    "westeurope": "West Europe",
    "japanwest": "Japan West",
    "japaneast": "Japan East",
    "brazilsouth": "Brazil South",
    "australiaeast": "Australia East",
    "australiasoutheast": "Australia Southeast",
    "southindia": "South India",
    "centralindia": "Central India",
    "westindia": "West India",
    "canadacentral": "Canada Central",
    "canadaeast": "Canada East",
    "uksouth": "UK South",
    "ukwest": "UK West",
    "westcentralus": "West Central US",
    "westus2": "West US 2",
    "koreacentral": "Korea Central",
    "koreasouth": "Korea South",
    "francecentral": "France Central",
    "francesouth": "France South",
    "australiacentral": "Australia Central",
    "australiacentral2": "Australia Central 2",
    "southafricanorth": "South Africa North",
    "southafricawest": "South Africa West",
}

AZURE_REGION_DISPLAY_MAP_SHORT = {
    "eastasia": "ea",
    "southeastasia": "sea",
    "centralus": "cus",
    "eastus": "eus",
    "eastus2": "eus2",
    "westus": "wus",
    "northcentralus": "ncus",
    "southcentralus": "scus",
    "northeurope": "ne",
    "westeurope": "we",
    "japanwest": "jw",
    "japaneast": "je",
    "brazilsouth": "bs",
    "australiaeast": "ae",
    "australiasoutheast": "ase",
    "southindia": "si",
    "centralindia": "ci",
    "westindia": "wi",
    "canadacentral": "cc",
    "canadaeast": "ce",
    "uksouth": "us",
    "ukwest": "uw",
    "westcentralus": "wc",
    "westus2": "wus2",
    "koreacentral": "kc",
    "koreasouth": "ks",
    "francecentral": "fc",
    "francesouth": "fs",
    "australiacentral": "ac",
    "australiacentral2": "ac2",
    "southafricanorth": "san",
    "southafricawest": "saw",
}


def azure_region_long(value):
    try:
        return AZURE_REGION_DISPLAY_MAP_LONG[value]
    except KeyError:
        raise


def azure_region_short(value):
    try:
        return AZURE_REGION_DISPLAY_MAP_SHORT[value]
    except KeyError:
        raise


def azure_frontend_endpoint_name(value: Endpoint):
    if value.azure and value.azure.internal_name:
        return render_tfvalue(value.azure.internal_name)
    return render_tfvalue(value.key)


def zone_name(value: str) -> str:
    value = utils.strip_protocol(value)
    return ".".join(value.split(".")[1:])


# Azure specific filters
def service_plan_resource_name(value: str) -> str:
    """Retreive the resource name for a Azure app service plan.

    The reason to make this conditional is because of backwards compatability;
    existing environments already have a `functionapp` resource. We want to keep that intact.
    """
    if value == "default":
        return "functionapps"
    return f"functionapps_{value}"


STORE_SUPPORTED_SCOPES = [
    "manage_orders",
    "manage_my_orders",
    "view_orders",
    "manage_customers",
    "view_customers",
    "manage_my_profile",
]


def render_commercetools_scopes(
    value: List[str], project_key: str, store_key: str = ""
):
    scopes = []
    for scope in value:
        scopes.append(f'"{scope}:{project_key}",')
        if store_key and scope in STORE_SUPPORTED_SCOPES:
            scopes.append(f'"{scope}:{project_key}:{store_key}",')

    return "[\n" + "".join(scopes) + "\n]"


def component_endpoint_name(
    component: Component,
    endpoint: Endpoint,
) -> str:
    """Take an component and an site-endpoint, and return a Terraform reference to an output.

    The endpoint might have a different name in the component itself based on the mappings
    """
    for component_key, ep_key in component.endpoints.items():
        if ep_key == endpoint.key:
            return component_key

    raise ValueError(f"Endpoint {endpoint.key} not found on {component.name}")



class MachError(Exception):
    pass


class UpdateError(MachError):
    pass


class ParseError(MachError):
    def __init__(self, msg, details=None):
        super().__init__(msg)
        self.details = details

    def __str__(self):
        result = super().__str__()
        details = self.details_pretty
        if details:
            result += f"{details}"

        return result

    @property
    def details_pretty(self):
        if not self.details or not isinstance(self.details, dict):
            return ""

        return self._pretty(self.details)

    def _pretty(self, data, *, indent=0):
        if not isinstance(data, dict):
            return data

        return "\n" + "\n".join(
            [
                f"{' ' * indent}{k}: {self._pretty(v, indent=indent+2)}"
                for k, v in data.items()
            ]
        )


class ValidationError(MachError):
    pass



from dataclasses import field

from dataclasses_json import config
from marshmallow import ValidationError, fields

# Define a none value as a custom dataclasses field so that
# null values get excluded in a dataclass dump
none = lambda: field(default=None, metadata=config(exclude=lambda x: x is None))  # type: ignore
default = lambda value: field(
    default_factory=lambda: value, metadata=config(exclude=lambda x: x == value)  # type: ignore
)
list_ = lambda: field(default_factory=list, metadata=config(exclude=lambda x: not x))  # type: ignore # noqa
dict_ = lambda: field(default_factory=dict, metadata=config(exclude=lambda x: not x))  # type: ignore # noqa


class EndpointsField(fields.Dict):
    def _serialize(self, value, attr, obj, **kwargs):
        result = {}
        for endpoint in value:
            if endpoint.contains_defaults:
                result[endpoint.key] = endpoint.url
            else:
                result[endpoint.key] = endpoint.to_dict()

        return super()._deserialize(result, attr, obj, **kwargs)

    def _deserialize(self, value, attr, data, **kwargs):
        from .sites import Endpoint

        value = super()._deserialize(value, attr, data, **kwargs)
        result = []
        for k, v in value.items():
            if isinstance(v, str):
                result.append(Endpoint(key=k, url=v))
            elif isinstance(v, dict):
                v["key"] = k
                result.append(Endpoint.schema(infer_missing=True).load(v))  # type: ignore
            else:
                raise ValidationError(f"Unexpected value found for endpoint {k}")

        return result



from dataclasses import dataclass
from typing import Optional

from dataclasses_json import dataclass_json
from dataclasses_jsonschema import JsonSchemaMixin

__all__ = ["MachComposerConfig"]


@dataclass_json
@dataclass
class MachComposerConfig(JsonSchemaMixin):
    version: str
    variables_file: Optional[str] = None



import concurrent.futures
import os
import re
import subprocess
from abc import ABC
from dataclasses import dataclass
from typing import List, Optional, Tuple, Union

import click
from mach import cache, exceptions, git, parse
from mach.types import ComponentConfig, MachConfig
from mach.variables import ignore_variable_not_found

NAME_RE = re.compile(r".* name: [\"']?(.*)[\"']?")
VERSION_RE = re.compile(r"(\s*version: )([\"']?.*[\"']?)")

ROOT_BLOCK_START = re.compile(r"^\w+")

MAX_WORKERS = 4

Updates = List[Tuple[ComponentConfig, str]]


@dataclass
class UpdaterInput:
    data: Union[MachConfig, List[ComponentConfig]]
    file: str

    @property
    def is_mach_config(self):
        return isinstance(self.data, MachConfig)

    @property
    def file_encrypted(self):
        return self.is_mach_config and self.data.file_encrypted


def update_file(
    file: str,
    *,
    component_name: Optional[str],
    new_version: Optional[str],
    verbose=False,
    check_only=False,
):
    try:
        with ignore_variable_not_found():
            config = parse.parse_config_from_file(file)
        data = UpdaterInput(config, file)
    except exceptions.ParseError as e:
        # We might have a components yml as input, try to parse that
        try:
            components = parse.parse_components(file)
        except exceptions.ParseError:
            # Raise original error
            raise e

        data = UpdaterInput(components, file)

    if component_name and new_version:
        update_config_component(data, component_name, new_version)
    else:
        update_config_components(data, verbose=verbose, check_only=check_only)


def update_config_component(  # noqa: C901
    updater_input: UpdaterInput,
    component_name: str,
    new_version: str,
):
    config = updater_input.data

    component = config.get_component(component_name)
    if not component:
        raise exceptions.MachError(f"Could not find component {component_name}")

    if component.version == new_version:
        click.echo(f"Component {component_name} is already on version {new_version}.")

    click.echo(f"Updating {component_name} to version {new_version}...")

    updater_cls = updater_for(updater_input)
    updater_cls.apply_updates(updater_input, [(component, new_version)])


def update_config_components(  # noqa: C901
    updater_input: UpdaterInput,
    *,
    check_only=False,
    verbose=False,
):
    """
    Update a given MACH configuration file.

    :param config: The MACH configuration to update components for
    :param check_only: Only check for updates; don't update the file
    :param verbose: Enable verbose output
    """
    intro_msg = f"Checking updates for components in {updater_input.file}"
    click.echo(intro_msg)
    click.echo("-" * len(intro_msg))

    updates: Updates = _fetch_changes(updater_input)

    if not check_only:
        updater_cls = updater_for(updater_input)
        updater_cls.apply_updates(updater_input, updates)


def updater_for(updater_input: UpdaterInput):
    if updater_input.is_mach_config:
        return SopsUpdater if updater_input.file_encrypted else FileUpdater

    return ComponentsFileUpdater


def _fetch_changes(updater_input: UpdaterInput) -> Updates:
    cache_dir = cache.cache_dir_for(updater_input.file)

    if updater_input.is_mach_config:
        components = updater_input.data.components
    else:
        components = updater_input.data

    updates: Updates = []

    def _inner_update(component):
        outputs = [f"Updates for {component.name}..."]

        match = re.match(r"^git::(.*)", component.source)
        if not match:
            outputs.append(
                f"  Cannot check {component.name} component since it doesn't have a Git source defined"  # noqa
            )
            click.echo("\n".join(outputs))
            return

        component_dir = os.path.join(cache_dir, component.name)
        repo = match.group(1)
        match = re.match(r"(.*\/.*)(\/\/.*)$", repo)
        if match:
            repo = match.group(1)
        git.ensure_local(repo, component_dir)

        commits = git.history(component_dir, component.version, branch=component.branch)
        if not commits:
            outputs.append("  No updates\n")
            click.echo("\n".join(outputs))
            return

        for commit in commits:
            outputs.append(f"  {commit.id}: {commit.msg} <{commit.author}>")

        outputs.append("")
        click.echo("\n".join(outputs))
        updates.append((component, commits[0].id))

    with concurrent.futures.ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:
        executor.map(_inner_update, components)

    return updates


class BaseUpdater(ABC):
    """Updater which update component version in-place.

    We'll use a very basic search-and-replace based on regular expressions
    instead of the yaml parser to not mess with any formatting.
    """

    @classmethod
    def apply_updates(cls, updater_input: UpdaterInput, updates: Updates):
        """Apply given updates to the file."""
        instance = cls()
        instance.apply(updater_input, updates)

    def apply(self, updater_input: UpdaterInput, updates: Updates):
        raise NotImplementedError()


class SopsUpdater(BaseUpdater):
    def apply(self, updater_input: UpdaterInput, updates: Updates):
        assert updater_input.file_encrypted

        if not updater_input.is_mach_config:
            raise NotImplementedError(
                "SOPS support not supported yet for updating a components file"
            )

        component_indexes = {
            c.name: i for i, c in enumerate(updater_input.data.components)
        }

        for component, version in updates:

            index = component_indexes[component.name]
            cmd = [
                "sops",
                "--set",
                f'["components"][{index}]["version"] "{version}"',
                updater_input.file,
            ]

            try:
                subprocess.run(cmd, check=True)
            except subprocess.CalledProcessError as e:
                click.echo(f"Failed to update {component.name}: {e}")
            else:
                click.echo(f"Updated {component.name} to {version} using SOPS")


class BaseFileUpdater(BaseUpdater):
    """Updater which update component version in-place.

    We'll use a very basic search-and-replace based on regular expressions
    instead of the yaml parser to not mess with any formatting.
    """

    def apply(self, updater_input: UpdaterInput, updates: Updates):
        click.echo("Writing updated to file...")
        file = updater_input.file
        self.current_component: Optional[ComponentConfig] = None
        self.updates = {component.name: version for component, version in updates}
        self.applied = []
        self.component_map = {component.name: component for component, _ in updates}

        with open(file) as f:
            lines = f.readlines()

        newlines = [self.process_line(line) for line in lines]

        not_applied = set(self.updates.keys()) - set(self.applied)
        if not_applied:
            click.echo(
                click.style(
                    f"Unable to apply all updates to the components: {', '.join(not_applied)}",
                    fg="yellow",
                )
            )

        with open(file, mode="w") as f:
            for line in newlines:
                f.write(line)

    def process_line(self, line: str) -> str:
        raise NotImplementedError()

    def process_component_line(self, line: str):
        name_match = NAME_RE.match(line)
        if name_match:
            component_name = name_match.group(1)

            try:
                self.current_component = self.component_map[component_name]
            except KeyError:
                self.current_component = None

            return line

        if not self.current_component:
            return line

        match = VERSION_RE.match(line)
        if not match:
            return line

        assert self.current_component.version in match.group(2)

        try:
            new_version = self.updates[self.current_component.name]
        except KeyError:
            return line

        new_version = f'"{new_version}"'
        self.applied.append(self.current_component.name)
        return VERSION_RE.sub(rf"\g<1>{new_version}", line)


class FileUpdater(BaseFileUpdater):
    def __init__(self):
        self.in_components = False

    def process_line(self, line: str) -> str:
        if line.startswith("components:"):
            self.in_components = True
        elif ROOT_BLOCK_START.match(line):
            self.in_components = False
        elif self.in_components:
            return self.process_component_line(line)

        return line


class ComponentsFileUpdater(BaseFileUpdater):
    def process_line(self, line: str) -> str:
        return self.process_component_line(line)



import re
from itertools import chain
from typing import Any, List

import click
from mach import types
from mach.__version__ import __version__
from mach.exceptions import ValidationError
from mach.variables import EmptyVar

STORE_KEY_RE = re.compile(r"^[\w_-]*$")


def check_variable(val: Any) -> bool:
    if type(val) == EmptyVar:
        return True
    return bool(val)


# Short-hand for readability
c = check_variable


def validate_config(config: types.MachConfig, *, ignore_version=True):
    """Check the config for invalid configuration."""
    if not ignore_version and config.mach_composer.version != __version__:
        raise ValidationError(
            f"MACH composer version defined in configuration ({config.mach_composer.version}) "
            f"does not match current version {__version__}"
        )

    validate_general_config(config.general_config)
    validate_components(config)

    for site in config.sites:
        validate_site(site, config=config)


def validate_general_config(config: types.GlobalConfig):
    if config.cloud == types.CloudOption.AZURE:
        if not config.azure:
            raise ValidationError("Missing azure configuration")
        if not config.terraform_config.azure_remote_state:
            raise ValidationError("Missing azure_remote_state configuration")
        if config.terraform_config.aws_remote_state:
            raise ValidationError(
                "Found aws_remote_state configuration, while cloud is set to 'azure'"
            )
    elif config.cloud == types.CloudOption.AWS:
        if not config.terraform_config.aws_remote_state:
            raise ValidationError("Missing aws_remote_state configuration")
        if config.terraform_config.azure_remote_state:
            raise ValidationError(
                "Found azure_remote_state configuration, while cloud is set to 'aws'"
            )

    if config.terraform_config:
        validate_terraform_config(config.terraform_config)

    if config.sentry:
        validate_sentry_config(config.sentry)


def validate_terraform_config(config: types.TerraformConfig):
    if config.providers:
        click.secho("Terraform provider versions", bold=True, fg="yellow")
        click.secho(
            "\n".join(
                [
                    "You are using custom Terraform provider versions.",
                    "Please be aware that some unexpected changes might occur compared to the MACH defaults.",  # noqa
                ]
            ),
            fg="yellow",
        )


def validate_site(site: types.Site, *, config: types.MachConfig):
    if config.general_config.cloud == types.CloudOption.AWS and not site.aws:
        raise ValidationError(f"Site {site.identifier} is missing an aws configuration")

    if config.general_config.cloud == types.CloudOption.AZURE:
        if site.endpoints and not site.azure.frontdoor:
            raise ValidationError(
                f"Site {site.identifier} has endpoints defined but frontdoor is not configured. "
                "Make sure your azure.frontdoor block is correcty configured in either the global "
                "or site configuration."
            )

    validate_endpoints(site, config.general_config.cloud)
    validate_commercetools(site)

    if site.components:
        validate_site_components(site.components, site=site)


def validate_endpoints(site: types.Site, cloud: types.CloudOption):
    # Construct lookup dictionary of all endpoints with the components that use them
    expected_endpoint_names = set(
        chain.from_iterable(c.endpoints.values() for c in site.components)
    )
    endpoint_names = {e.key for e in site.endpoints}

    missing = expected_endpoint_names - endpoint_names
    if missing:
        raise ValidationError(f"Missing required endpoints {', '.join(missing)}")


def validate_site_components(components: List[types.Component], *, site: types.Site):
    """Sanity checks on component configuration per site."""
    defined_stores = (
        [s.key for s in site.commercetools.stores] if site.commercetools else []
    )

    for component in components:
        if component.health_check_path and not component.health_check_path.startswith(
            "/"
        ):
            raise ValidationError(
                f"Component health check {component.health_check_path} does "
                "not start with '/'."
            )

        for store in component.store_variables.keys():
            if store not in defined_stores:
                raise ValidationError(
                    f"Store {store} is not defined in your commercetools stores definition"
                )

        if site.azure:
            service_plans = site.azure.service_plans
            if (
                component.azure
                and component.azure.service_plan
                and component.azure.service_plan not in service_plans
            ):
                raise ValidationError(
                    f"Component {component.name} requires service plan "
                    f"{component.azure.service_plan} which is not defined in the "
                    "Azure configuration."
                )


def validate_commercetools(site: types.Site):
    if site.commercetools:
        validate_store_keys(site.commercetools)
        validate_taxes(site.commercetools)


def validate_store_keys(ct_settings: types.CommercetoolsSettings):
    """Sanity checks on store values."""
    if ct_settings.stores:
        store_keys = [store.key for store in ct_settings.stores]
        for key in store_keys:
            if len(key) < 2:
                raise ValidationError(
                    f"Store key {key} should be minimum two characters."
                )
            if store_keys.count(key) != 1:
                raise ValidationError(f"Store key {key} must be unique.")

            if not STORE_KEY_RE.match(key):
                raise ValidationError(
                    f"Store key {key} may only contain alphanumeric characters or underscores"
                )


def validate_taxes(ct_settings: types.CommercetoolsSettings):
    """Ensure either `taxes` or `tax_categories` are used."""
    if ct_settings.taxes and ct_settings.tax_categories:
        raise ValidationError("`taxes` and `tax_categories` are mutually exclusive.")


def validate_components(config: types.MachConfig):
    """Validate global component data is valid."""
    if config.general_config.cloud == types.CloudOption.AWS:
        validate_aws_components(config)
    elif config.general_config.cloud == types.CloudOption.AZURE:
        validate_azure_components(config)


def validate_sentry_config(config: types.SentryConfig):
    if not any([c(config.dsn), c(config.auth_token)]):
        raise ValidationError("sentry: Either dsn or auth_token should be set")

    if all([c(config.dsn), c(config.auth_token)]):
        raise ValidationError("sentry: Only a dsn or auth_token should be defined")

    if config.auth_token and not any([c(config.project), c(config.organization)]):
        raise ValidationError(
            "sentry: A project and organization should be defined when using an auth_token"
        )


def validate_aws_components(config: types.MachConfig):
    """Validate components specifically for AWS usage."""
    pass


def validate_azure_components(config: types.MachConfig):
    """Validate components specifically for Azure usage.

    Only requirements for now is that a correct short_name should be set.
    Otherwise problems will arise when creating the Azure resources since
    for example Storage Accounts names have a limited length.
    """
    for comp in config.components:
        if comp.artifacts:
            raise ValidationError(
                f"The artifacts options on the '{comp.name}' component "
                "are not supported when targetting Azure"
            )

        if "azure" not in comp.integrations:
            continue

        # azure naming length is limited, so verify it doesn't get too long.
        assert comp.azure and comp.azure.short_name  # Should have been set by parser
        if len(comp.azure.short_name) > 10:
            raise ValidationError(
                f"Component {comp.name} short_name '{comp.azure.short_name}' "
                "cannot be more than 10 characters."
            )



from dataclasses import dataclass
from enum import Enum
from typing import Dict, Optional

from dataclasses_json import dataclass_json
from dataclasses_jsonschema import JsonSchemaMixin

from . import fields
from .shared import ServicePlan

__all__ = [
    "AWSTFState",
    "AmplienceConfig",
    "AzureConfig",
    "AzureTFState",
    "CloudOption",
    "ContentfulConfig",
    "FrontdoorSettings",
    "GlobalConfig",
    "SentryConfig",
    "TerraformConfig",
    "TerraformProviders",
]


class StringEnum(str, Enum):
    def __eq__(self, value):
        if isinstance(value, str):
            return self.value == value
        return super().__eq__(value)

    def __ne__(self, value):
        if isinstance(value, str):
            return self.value != value
        return super().__ne__(value)


@dataclass_json
@dataclass
class AzureTFState(JsonSchemaMixin):
    """Azure storage account state backend configuration."""

    resource_group: str
    storage_account: str
    container_name: str
    state_folder: str


@dataclass_json
@dataclass
class AWSTFState(JsonSchemaMixin):
    """AWS S3 bucket state backend configuration."""

    bucket: str
    key_prefix: str
    region: str
    role_arn: Optional[str] = fields.none()
    lock_table: Optional[str] = fields.none()
    encrypt = True


@dataclass_json
@dataclass
class TerraformProviders(JsonSchemaMixin):
    """Terraform provider version overwrites."""

    aws: Optional[str] = fields.none()
    azure: Optional[str] = fields.none()
    commercetools: Optional[str] = fields.none()
    sentry: Optional[str] = fields.none()
    contentful: Optional[str] = fields.none()
    amplience: Optional[str] = fields.none()


@dataclass_json
@dataclass
class TerraformConfig(JsonSchemaMixin):
    """Terraform configuration."""

    azure_remote_state: Optional[AzureTFState] = fields.none()
    aws_remote_state: Optional[AWSTFState] = fields.none()
    providers: Optional[TerraformProviders] = fields.none()


@dataclass_json
@dataclass
class SentryConfig(JsonSchemaMixin):
    """Global Sentry configuration."""

    dsn: Optional[str] = fields.none()
    rate_limit_window: Optional[int] = fields.none()
    rate_limit_count: Optional[int] = fields.none()

    auth_token: Optional[str] = fields.none()
    base_url: Optional[str] = fields.none()
    project: Optional[str] = fields.none()
    organization: Optional[str] = fields.none()

    @property
    def managed(self):
        """Indicate if the Sentry DSN should be managed by MACH."""
        return bool(self.auth_token)


@dataclass_json
@dataclass
class FrontdoorSslConfig(JsonSchemaMixin):
    name: str
    resource_group: str
    secret_name: str


@dataclass_json
@dataclass
class FrontdoorSettings(JsonSchemaMixin):
    """Frontdoor settings."""

    dns_resource_group: str
    ssl_key_vault: Optional[FrontdoorSslConfig] = fields.none()
    # Undocumented option to workaround some tenacious issues
    # with using Frontdoor in the Azure Terraform provider
    suppress_changes: bool = fields.default(False)


@dataclass_json
@dataclass
class AzureConfig(JsonSchemaMixin):
    """Azure configuration."""

    tenant_id: str
    subscription_id: str
    region: str
    frontdoor: Optional[FrontdoorSettings] = fields.none()
    resources_prefix: str = ""
    service_object_ids: Dict[str, str] = fields.dict_()
    service_plans: Dict[str, ServicePlan] = fields.dict_()


@dataclass_json
@dataclass
class ContentfulConfig(JsonSchemaMixin):
    """Generic Contenful configuration."""

    cma_token: str
    organization_id: str


@dataclass_json
@dataclass
class AmplienceConfig(JsonSchemaMixin):
    """Generic Amplience configuration."""

    client_id: str
    client_secret: str


class CloudOption(StringEnum):
    AWS = "aws"
    AZURE = "azure"


@dataclass_json
@dataclass
class GlobalConfig(JsonSchemaMixin):
    """Config that is shared across sites."""

    environment: str
    terraform_config: TerraformConfig
    cloud: CloudOption
    sentry: Optional[SentryConfig] = fields.none()
    azure: Optional[AzureConfig] = fields.none()
    contentful: Optional[ContentfulConfig] = fields.none()
    amplience: Optional[AmplienceConfig] = fields.none()



import glob
import subprocess
import sys
import textwrap
from functools import update_wrapper
from typing import List, Optional

import click
from mach import bootstrap as _bootstrap
from mach import git, parse, updater
from mach.__version__ import __version__
from mach.build import build_packages
from mach.exceptions import MachError, ParseError
from mach.terraform import (
    apply_terraform,
    generate_terraform,
    init_terraform,
    plan_terraform,
)
from mach.variables import ignore_variable_not_found


class Command(click.Command):
    def invoke(self, ctx):
        try:
            return super().invoke(ctx)
        except MachError as e:
            raise click.ClickException(str(e)) from e


class Group(click.Group):
    def command(self, *args, **kwargs):
        kwargs.setdefault("cls", Command)
        return super().command(*args, **kwargs)


@click.group(cls=Group)
@click.version_option(__version__)
def mach():
    pass


def terraform_command(f):
    @click.option(
        "-f",
        "--file",
        default=None,
        help="YAML file to parse. If not set parse all *.yml files.",
    )
    @click.option(
        "-s",
        "--site",
        default=None,
        help="Site to parse. If not set parse all sites.",
    )
    @click.option(
        "--ignore-version",
        is_flag=True,
        default=False,
        help="Skip MACH composer version check",
    )
    @click.option(
        "--output-path",
        default="deployments",
        help="Output path, defaults to `cwd`/deployments.",
    )
    @click.option(
        "--var-file",
        help="Use a variable file to parse the configuration with",
    )
    def new_func(
        file, site, output_path: str, ignore_version: bool, var_file, **kwargs
    ):
        files = get_input_files(file, var_file=var_file)

        configs = parse.parse_configs(
            files, output_path, ignore_version=ignore_version, var_file=var_file
        )

        if not configs:
            raise click.ClickException("No valid MACH configurations found")

        try:
            result = f(
                file=file,
                site=site,
                configs=configs,
                **kwargs,
            )
        except subprocess.CalledProcessError as e:
            click.echo("Failed to run")
            sys.exit(e.returncode)
        else:
            click.echo("Done 👍")
            return result

    return update_wrapper(new_func, f)


@mach.command()
@terraform_command
def generate(file, site, configs, *args, **kwargs):
    """Generate the Terraform files."""
    for config in configs:
        generate_terraform(config, site=site)


@mach.command()
@terraform_command
def init(file, site, configs, *args, **kwargs):
    """Initialize site directories Terraform files."""
    for config in configs:
        generate_terraform(config, site=site)
        init_terraform(config, site=site)


@mach.command()
@click.option(
    "--with-sp-login",
    is_flag=True,
    default=False,
    help="If az login with service principal environment variables "
    "(ARM_CLIENT_ID, ARM_CLIENT_SECRET, ARM_TENANT_ID) should be done.",
)
@click.option(
    "-c",
    "--component",
    multiple=True,
    default=[],
    help="",
)
@click.option(
    "--reuse",
    default=False,
    is_flag=True,
    help="Supress a terraform init for improved speed (not recommended for production usage)",
)
@click.option(
    "--destroy",
    default=False,
    is_flag=True,
    help="Destroy option is a convenient way to destroy all remote objects managed by this mach config",  # noqa
)
@terraform_command
def plan(
    file, site, configs, with_sp_login, component, reuse, destroy, *args, **kwargs
):
    """Output the deploy plan."""
    for config in configs:
        generate_terraform(config, site=site)
        plan_terraform(
            config,
            site=site,
            components=component,
            with_sp_login=with_sp_login,
            reuse=reuse,
            destroy=destroy,
        )


@mach.command()
@click.option(
    "--with-sp-login",
    is_flag=True,
    default=False,
    help="If az login with service principal environment variables "
    "(ARM_CLIENT_ID, ARM_CLIENT_SECRET, ARM_TENANT_ID) should be done.",
)
@click.option(
    "--auto-approve",
    is_flag=True,
    default=False,
    help="",
)
@click.option(
    "-c",
    "--component",
    multiple=True,
    default=[],
    help="",
)
@click.option(
    "--reuse",
    is_flag=True,
    default=False,
    help="Supress a terraform init for improved speed (not recommended for production usage)",
)
@click.option(
    "--destroy",
    default=False,
    is_flag=True,
    help="Destroy option is a convenient way to destroy all remote objects managed by this mach config",  # noqa
)
@terraform_command
def apply(
    file,
    site,
    configs,
    with_sp_login,
    auto_approve,
    component,
    reuse,
    destroy,
    *args,
    **kwargs,
):
    """Apply the configuration."""
    for config in configs:
        build_packages(config, restrict_components=component)
        generate_terraform(config, site=site)
        apply_terraform(
            config,
            site=site,
            components=component,
            with_sp_login=with_sp_login,
            auto_approve=auto_approve,
            reuse=reuse,
            destroy=destroy,
        )


# flake8: noqa: C901
@mach.command()
@click.option(
    "-f",
    "--file",
    default=None,
    help="YAML file to update. If not set update all *.yml files.",
)
@click.option("-v", "--verbose", is_flag=True, default=False, help="Verbose output.")
@click.option(
    "--check",
    default=False,
    is_flag=True,
    help="Only checks for updates, doesnt change files.",
)
@click.option(
    "-c",
    "--commit",
    default=False,
    is_flag=True,
    help="Automatically commits the change.",
)
@click.option(
    "-m", "--commit-message", default=None, help="Use a custom message for the commit."
)
@click.argument("component", required=False)
@click.argument("version", required=False)
def update(
    file: str,
    check: bool,
    verbose: bool,
    component: str,
    version: str,
    commit: bool,
    commit_message: str,
):
    """Update all (or a given) component.

    When no component and version is given, it will check the git repositories for any updates.
    This command can also be used to manually update a single component by specifying a component
    and version.
    """
    if check and commit:
        raise click.ClickException(
            "check_only is not possible when create_commit is enabled."
        )

    if component and not version:
        if "@" not in component:
            raise click.ClickException(
                f"When specifying a component ({component}) you should specify a version as well"
            )
        component, version = component.split("@")

    for file in get_input_files(file):
        try:
            updater.update_file(
                file,
                component_name=component,
                new_version=version,
                verbose=verbose,
                check_only=check,
            )
        except ParseError as e:
            click.echo(textwrap.indent(str(e), "  "))
            continue

        if commit:
            git.add(file)

    if commit:
        if commit_message:
            commit_msg = commit_message
        elif component:
            commit_msg = f"Updated {component} component"
        else:
            commit_msg = "Updated components"

        git.commit(commit_msg)


@mach.command()
@click.option(
    "-f",
    "--file",
    default=None,
    help="YAML file to read. If not set read all *.yml files.",
)
def components(file: str):
    """List all components."""
    files = get_input_files(file)

    with ignore_variable_not_found():
        configs = parse.parse_configs(files)

    for config in configs:
        click.echo(f"{config.file}:")
        for component in config.components:
            click.echo(f" - {component.name}")
            click.echo(f"   version: {component.version}")

        click.echo("")


@mach.command()
@click.option(
    "-f",
    "--file",
    default=None,
    help="YAML file to read. If not set read all *.yml files.",
)
def sites(file: str):
    """List all sites."""
    files = get_input_files(file)

    with ignore_variable_not_found():
        configs = parse.parse_configs(files)

    for config in configs:
        click.echo(f"{config.file}:")
        for site in config.sites:
            click.echo(f" - {site.identifier}")
            click.echo("   components:")
            for component in site.components:
                click.echo(f"     {component.name}")

        click.echo("")


@mach.command()
@click.option(
    "-o",
    "--output",
    help="Output file or directory.",
)
@click.option(
    "-c",
    "--cookiecutter",
    default="",
    help="cookiecutter repository to generate from.",
)
@click.argument("type_", required=True, type=click.Choice(["config", "component"]))
def bootstrap(output: str, type_: str, cookiecutter: str):
    """Bootstraps a configuration or component."""
    if type_ == "config":
        _bootstrap.create_configuration(output or "main.yml")
    if type_ == "component":
        _bootstrap.create_component(output, cookiecutter)


@mach.command()
@click.argument("site", required=True)
@click.option(
    "-f",
    "--file",
    default=None,
    help="YAML file to read. If not set read all *.yml files.",
)
def path(file: str, site: str):
    """Return output path for given site."""
    # Suppress click output, we only want to output the output path
    click.echo = lambda *args, **kwargs: None
    files = get_input_files(file)
    configs = parse.parse_configs(files, "deployments")
    for config in configs:
        click.echo(f"{config.file}:")
        for site_ in config.sites:
            if site_.identifier == site:
                sys.stdout.write(f"{config.deployment_path_for(site_)}\n")
                return
    sys.exit(f"Could not find site definition {site}\n")


def get_input_files(file: Optional[str], *, var_file: str = None) -> List[str]:
    """Determine input files. If file is not specified use all *.yml files."""
    if file:
        files = [file]
    else:
        files = glob.glob("./*.yml")
    if not files:
        click.echo("No .yml files found")
        sys.exit(1)

    # If a var-file is given, strip it from the list of files to parse a MACH configurations.
    # This is mainly a convenience for when you have the following files;
    # - main.yml
    # - variables.yml
    # and to run `mach apply --var-file variables.yml`
    # instead of `mach apply -f main.yml --var-file variables.yml`
    if var_file:
        files = filter(lambda f: f.lstrip("./") != var_file, files)
    return files
